--- a/drivers/net/wireless/ralink/rt2x00/rt2800lib.c
+++ b/drivers/net/wireless/ralink/rt2x00/rt2800lib.c
@@ -1846,9 +1846,31 @@
 	rt2800_register_write(rt2x00dev, GF40_PROT_CFG, reg);
 }
 
+static void rt2800_update_timing(struct rt2x00_dev *rt2x00dev)
+{
+	struct rt2800_drv_data *drv_data = rt2x00dev->drv_data;
+	int slot_time = drv_data->slot_time;
+	int timeout;
+	u32 reg;
+
+	slot_time += 3 * drv_data->coverage_class;
+
+	reg = rt2800_register_read(rt2x00dev, BKOFF_SLOT_CFG);
+	rt2x00_set_field32(&reg, BKOFF_SLOT_CFG_SLOT_TIME, slot_time);
+	rt2800_register_write(rt2x00dev, BKOFF_SLOT_CFG, reg);
+
+	reg = rt2800_register_read(rt2x00dev, XIFS_TIME_CFG);
+	timeout = slot_time + rt2x00_get_field32(reg, XIFS_TIME_CFG_EIFS);
+
+	reg = rt2800_register_read(rt2x00dev, TX_TIMEOUT_CFG);
+	rt2x00_set_field32(&reg, TX_TIMEOUT_CFG_RX_ACK_TIMEOUT, timeout);
+	rt2800_register_write(rt2x00dev, TX_TIMEOUT_CFG, reg);
+}
+
 void rt2800_config_erp(struct rt2x00_dev *rt2x00dev, struct rt2x00lib_erp *erp,
 		       u32 changed)
 {
+	struct rt2800_drv_data *drv_data = rt2x00dev->drv_data;
 	u32 reg;
 
 	if (changed & BSS_CHANGED_ERP_PREAMBLE) {
@@ -1872,14 +1894,13 @@
 	}
 
 	if (changed & BSS_CHANGED_ERP_SLOT) {
-		reg = rt2800_register_read(rt2x00dev, BKOFF_SLOT_CFG);
-		rt2x00_set_field32(&reg, BKOFF_SLOT_CFG_SLOT_TIME,
-				   erp->slot_time);
-		rt2800_register_write(rt2x00dev, BKOFF_SLOT_CFG, reg);
-
 		reg = rt2800_register_read(rt2x00dev, XIFS_TIME_CFG);
 		rt2x00_set_field32(&reg, XIFS_TIME_CFG_EIFS, erp->eifs);
 		rt2800_register_write(rt2x00dev, XIFS_TIME_CFG, reg);
+                
+		drv_data->slot_time = erp->slot_time;
+		rt2800_update_timing(rt2x00dev);
+                
 	}
 
 	if (changed & BSS_CHANGED_BEACON_INT) {
@@ -1887,6 +1908,7 @@
 		rt2x00_set_field32(&reg, BCN_TIME_CFG_BEACON_INTERVAL,
 				   erp->beacon_int * 16);
 		rt2800_register_write(rt2x00dev, BCN_TIME_CFG, reg);
+                rt2800_update_timing(rt2x00dev);
 	}
 
 	if (changed & BSS_CHANGED_HT)
@@ -10347,6 +10369,16 @@
 }
 EXPORT_SYMBOL_GPL(rt2800_get_survey);
 
+void rt2800_set_coverage_class(struct ieee80211_hw *hw, s16 cc)
+{
+	struct rt2x00_dev *rt2x00dev = hw->priv;
+	struct rt2800_drv_data *drv_data = rt2x00dev->drv_data;
+
+	drv_data->coverage_class = cc;
+	rt2800_update_timing(rt2x00dev);
+}
+EXPORT_SYMBOL_GPL(rt2800_set_coverage_class);
+
 MODULE_AUTHOR(DRV_PROJECT ", Bartlomiej Zolnierkiewicz");
 MODULE_VERSION(DRV_VERSION);
 MODULE_DESCRIPTION("Ralink RT2800 library");
--- a/drivers/net/wireless/ralink/rt2x00/rt2800lib.h
+++ b/drivers/net/wireless/ralink/rt2x00/rt2800lib.h
@@ -42,6 +42,8 @@
 	u8 txmixer_gain_24g;
 	u8 txmixer_gain_5g;
 	u8 max_psdu;
+        int slot_time;
+	int coverage_class;
 	unsigned int tbtt_tick;
 	unsigned int ampdu_factor_cnt[4];
 	DECLARE_BITMAP(sta_ids, STA_IDS_SIZE);
@@ -265,4 +267,6 @@
 			       unsigned short *txwi_size,
 			       unsigned short *rxwi_size);
 
+void rt2800_set_coverage_class(struct ieee80211_hw *hw, s16 cc);
+
 #endif /* RT2800LIB_H */
--- a/drivers/net/wireless/ralink/rt2x00/rt2800usb.c
+++ b/drivers/net/wireless/ralink/rt2x00/rt2800usb.c
@@ -808,6 +808,7 @@
 	.get_survey		= rt2800_get_survey,
 	.get_ringparam		= rt2x00mac_get_ringparam,
 	.tx_frames_pending	= rt2x00mac_tx_frames_pending,
+	.set_coverage_class	= rt2800_set_coverage_class,
 };
 
 static const struct rt2800_ops rt2800usb_rt2800_ops = {
--- a/drivers/net/wireless/ralink/rt2x00/rt2800pci.c
+++ b/drivers/net/wireless/ralink/rt2x00/rt2800pci.c
@@ -322,6 +322,7 @@ static const struct ieee80211_ops rt2800
 	.get_survey		= rt2800_get_survey,
 	.get_ringparam		= rt2x00mac_get_ringparam,
 	.tx_frames_pending	= rt2x00mac_tx_frames_pending,
+	.set_coverage_class	= rt2800_set_coverage_class,
 };
 
 static const struct rt2800_ops rt2800pci_rt2800_ops = {
--- a/drivers/net/wireless/ralink/rt2x00/rt2800soc.c
+++ b/drivers/net/wireless/ralink/rt2x00/rt2800soc.c
@@ -155,6 +155,7 @@ static const struct ieee80211_ops rt2800
 	.get_survey		= rt2800_get_survey,
 	.get_ringparam		= rt2x00mac_get_ringparam,
 	.tx_frames_pending	= rt2x00mac_tx_frames_pending,
+	.set_coverage_class	= rt2800_set_coverage_class,
 };
 
 static const struct rt2800_ops rt2800soc_rt2800_ops = {
